/* Copyright (C) 2006, 2013  Stephan Kreutzer
 *
 * This file is part of "Random Access File Example".
 *
 * "Random Access File Example" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 or any later version,
 * as published by the Free Software Foundation.
 *
 * "Random Access File Example" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License 3 for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with "Random Access File Example". If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * @file $/ra_file.cpp
 * @brief Implementation file of $/ra_file.h.
 * @since 2006-11-08
 * @author Stephan Kreutzer
 */



#include "ra_file.h"



/**
 * @class ra_file
 */

ra_file::ra_file():
fp(NULL)
{

}

ra_file::ra_file(const ra_file& rhs):
fp(NULL)
{

}

ra_file::~ra_file()
{
    if (ra_file::fp != NULL)
    {
        std::fclose(ra_file::fp);
    }
}

/**
 * @param[in] path A string containing the path of a file which shall be
 *     opened by this method.
 * @param[in] open_mode A string containing the open mode options (see the
 *     std::fopen() parameters).
 * @retval true An error has occurred.
 */
bool ra_file::open(const char* path, const char* open_mode)
{
    if (ra_file::fp != NULL)
    {
        // A file is already opened, so it is
        // important to close it first.
        if (std::fclose(ra_file::fp) != 0)
        {
            // Can't close file!
            return true;
        }
        else
        {
            // File closed successfully. Set
            // "fp" to NULL to indicate
            // that the file pointer is not
            // linked to a file any more.
            ra_file::fp = NULL;
        }
    }

    if ((ra_file::fp = std::fopen(&path[0], open_mode)) == NULL)
    {
        // Can't open file!
        return true;
    }
    else
    {
        // File opened successfully.
        return false;
    }
}

/**
 * @retval true An error has occurred.
 */
bool ra_file::close()
{
    if (ra_file::fp != NULL)
    {
        if (std::fclose(ra_file::fp) != 0)
        {
            // Can't close file!
            return true;
        }
        else
        {
            // File closed successfully. Set
            // "fp" to NULL to indicate
            // that the file pointer is not
            // linked to a file any more.
            ra_file::fp = NULL;
            return false;
        }
    }
    else
    {
        // Can't close file because none was
        // opened...but this isn't an error
        // so the return value is "false".
        return false;
    }
}

/**
 * @brief This method gives information about the status of the file pointer.
 * @retval false There's currently no file opened.
 * @retval true There's currently a file opened.
 */
bool ra_file::raf_opened()
{
    if (ra_file::fp == NULL)
    {
        return false;
    }
    
    return true;
}

/**
 * @retval -1 An error has occurred. 
 * @return Returns the file size in bytes.
 */
long ra_file::raf_size()
{
    if (ra_file::fp != NULL)
    {
        long file_size = 0;

        // Get file size.
        std::fseek(ra_file::fp, 0, SEEK_END);
        file_size = std::ftell(ra_file::fp);
        // Reset file pointer.
        std::fseek(ra_file::fp, 0, SEEK_SET);

        return file_size;
    }
    else
    {
        // Can't get file size, no file
        // opened!
        return -1;
    }
}

/**
 * @brief This method is able to read structured data from a file.
 * @details One can choose a data block, select a data field within such a
 *     block and read as much bytes as he wants.
 * @param[out] destination ra_file::raf_read() will write the read bytes into
 *     this buffer (should be at least as large as the size specified by \p n.
 * @param[in] block_offset Specifies the data block which shall be read.
 * @param[in] block_size This parameter takes the size of one data block.
 * @param[in] field_offset This parameter specifies how much bytes ra_file::raf_read()
 *     has to move from the begin of the chosen data block. So it is possible to
 *     select a field within the data block. The method will find the first byte
 *     it has to read at this position.
 * @param[in] n The amount of bytes ra_file::raf_read() shall read. The argument
 *     shouldn't be higher than the size of the destination (see \p destination).
 * @retval NULL An error has occurred.
 * @return \p destination.
 */
void* ra_file::raf_read(void* destination,
                        unsigned long block_offset,
                        unsigned long block_size,
                        unsigned long field_offset,
                        std::size_t n)
{
    if (ra_file::fp == NULL)
    {
        // Can't read from file, no file
        // opened!
        return NULL;
    }

    if (destination == NULL)
    {
        // 'destination' is not set to a valid address!
        return NULL;
    }


    long file_size = 0;

    if ((file_size = ra_file::raf_size()) == -1)
    {
        // Error: Can't get file size!
        return NULL;
    }

    if ((block_offset * block_size) + (field_offset + n) > static_cast<unsigned long> (file_size))
    {
        // The arguments would force raf_read() method to read
        // bytes after EOF!
        return NULL;
    }

    // Set file pointer to the first position where this
    // method shall begin with reading.
    std::fseek(ra_file::fp,
               (block_offset * block_size) + field_offset,
               SEEK_SET);

    if (std::fread(destination, 1, n, ra_file::fp) != n)
    {
        // fread() wasn't able to read the complete
        // amount of bytes - it seems like EOF is reached...
        return NULL;
    }
    else
    {
        // Method call was successful.
        std::fflush(ra_file::fp);
        std::fseek(ra_file::fp, 0, SEEK_SET);
        return destination;
    }
}

/**
 * @brief The task of this method is to overwrite existing data blocks. If one
 *     wants to add a new data block to the file, he should better use ra_file::raf_add().
 * @param[in] source The source of the new data. This method will copy the
 *     bytes stored at \p source into the file.
 * @param[in] block_offset Specifies the data block from which this method shall
 *     write into.
 * @param[in] block_size This parameter takes the size of one data block.
 * @param[in] field_offset This parameter specifies how much bytes ra_file::raf_write()
 *     has to move from the begin of the chosen data block. So it is possible to
 *     select a field within the data block. The method will begin to write
 *     from this position on.
 * @param[in] n The amount of bytes ra_file::raf_write() shall copy from
 *     \p source into the file.
 * @retval true An error has occurred.
 */
bool ra_file::raf_write(const void* source,
                        unsigned long block_offset,
                        unsigned long block_size,
                        unsigned long field_offset,
                        std::size_t n)
{
    if (ra_file::fp == NULL)
    {
        // Can't write into file, no file
        // opened!
        return true;
    }

    if (source == NULL)
    {
        // 'source' is not set to a valid address!
        return true;
    }

    if ((field_offset + n) > block_size)
    {
        // Access violation! The arguments
        // would force raf_write() method to write
        // over the limits of a data block!
        return true;
    }


    long file_size = 0;

    if ((file_size = ra_file::raf_size()) == -1)
    {
        // Error: Can't get file size!
        return true;
    }

    if ((block_offset * block_size) + (field_offset + n) > static_cast<unsigned long> (file_size))
    {
        // The arguments would force raf_write() method to write
        // bytes after EOF! You should use ra_file::raf_add() method
        // to insert a new data block. ra_file::raf_write() is only for
        // editing existing data blocks.
        return true;
    }

    // Set file pointer to the first position where this
    // method shall begin with writing
    std::fseek(ra_file::fp,
               (block_offset * block_size) + field_offset,
               SEEK_SET);

    if (std::fwrite(source, 1, n, ra_file::fp) != n)
    {
        // fwrite() wasn't able to write the complete
        // amount of bytes - it seems like EOF is reached...
        return true;
    }
    else
    {
        // Method call was successful.
        std::fflush(ra_file::fp);
        std::fseek(ra_file::fp, 0, SEEK_SET);
        return false;
    }
}

/**
 * @brief This method adds/appends a new data block to the file.
 * @detail  This overloaded version will add an empty data block (empty means:
 *     all bytes of the data block will be '\0').
 * @param[in] block_size This parameter takes the size of one data block.
 * @retval true An error has occurred.
 */
bool ra_file::raf_add(unsigned long block_size)
{
    if (ra_file::fp == NULL)
    {
        // Can't add to file, no file
        // opened!
        return true;
    }

    // Set file pointer to EOF.
    std::fseek(ra_file::fp, 0, SEEK_END);

    // Write '\0'-characters into the file till a complete
    // data block is written.
    for (std::size_t i = 0; i < block_size; i++)
    {
        if (fputc('\0', ra_file::fp) == EOF)
        {
            // fputc() wasn't able to write another '\0'-character -
            // a write error has occurred.
            return true;
        }
    }

    std::fflush(ra_file::fp);
    std::fseek(ra_file::fp, 0, SEEK_SET);
    // Method call was successful.
    return false;
}

/**
 * @brief This method adds/appends a new data block to the file.
 * @details This overloaded version won't just add an empty data block, it will
 *     copy the bytes at \p source into the file. ra_file::raf_add() will also
 *     fill the data block with '\0', if \p source can't provide enought bytes
 *     to cover the whole data block.
 * @param[in] source The source of new data. This method will copy the bytes
 *     stored at \p source into the file (at the position specified by the
 *     overgiven arguments).
 * @param[in] block_size This parameter takes the size of one data block.
 * @param[in] field_offset This parameter specifies how much bytes ra_file::raf_add()
 *     has to move from the begin of the chosen data block. So it is possible to
 *     select a field within the data block. The method will begin to copy the
 *     bytes stored at \p source from this position on. All bytes before this
 *     position will be filled with '\0'.
 * @param[in] n  The amount of bytes ra_file::raf_add() shall copy from \p source
 *     into the file.
 * @retval true An error has occurred.
 */
bool ra_file::raf_add(const void* source,
                      unsigned long block_size,
                      unsigned long field_offset,
                      std::size_t n)
{
    if (ra_file::fp == NULL)
    {
        // Can't append to file, no file
        // opened!
        return true;
    }

    if (source == NULL)
    {
        // 'source' is not set to a valid address!
        return true;
    }

    if ((field_offset + n) > block_size)
    {
        // Access violation! The arguments
        // would force raf_add() method to write
        // over the limits of a data block!
        return true;
    }

    // Set file pointer to EOF.
    std::fseek(ra_file::fp, 0, SEEK_END);

    // Write '\0'-characters into the file till the
    // selected field is reached.
    for (std::size_t i = 0; i < field_offset; i++)
    {
        if (fputc('\0', ra_file::fp) == EOF)
        {
            // fputc() wasn't able to write another '\0'-character -
            // a write error has occurred.
            return true;
        }
    }

    if (std::fwrite(source, 1, n, ra_file::fp) != n)
    {
        // fwrite() wasn't able to write the complete
        // amount of bytes - it seems like EOF is reached...
        return true;
    }

    // Fill with '\0' to complete the block
    for (std::size_t i = 0; i < block_size - (field_offset + n); i++)
    {
        if (fputc('\0', ra_file::fp) == EOF)
        {
            // fputc() wasn't able to write another '\0'-character -
            // a write error has occurred.
            return true;
        }
    }

    std::fflush(ra_file::fp);
    std::fseek(ra_file::fp, 0, SEEK_SET);
    // Method call was successful.
    return false;
}



